function [ cdata Z ] = agglomerate_corres_v12( param, cdata ) 
% function to agglomerate correspondences 
% using the adaptive partial linkage model of the ICCV 2009 paper
% note that seed indexes is the same as match indexes

% 9 Nov 2010: upgrated to enforce the one-to-one constraint inside each clsuter, added by Minsu Cho, 
%             journal submitted version 

bPair = cdata.bPair;
clusterInfo = cdata.clusterInfo;

cdata.affinityMatrix(1:size(cdata.affinityMatrix,1)+1:end) = 0;
cdata.overlapMatrix(1:size(cdata.overlapMatrix,1)+1:end) = 0;
cur_dist_mat = squareform(cdata.affinityMatrix);
cur_mindist_mat = cur_dist_mat;
cur_ovlp_mat = squareform(cdata.overlapMatrix);
nCluster = size(cdata.affinityMatrix,1);

if param.bOneToOne % induce the matching constraint on the dissim matrix
    cur_dist_mat( find(cur_ovlp_mat) ) = Inf;
    cur_mindist_mat( find(cur_ovlp_mat) ) = Inf;
end

m = nCluster;
Z = zeros(m-1,3); % allocate the output matrix.

% during updating clusters, cluster index is constantly changing, R is
% a index vector mapping the original index to the current (row)
% index in Y.  N denotes how many points are contained in each cluster.
N = zeros(1,2*nCluster-1);
N(1:m) = 1;
n = m; % since m is changing, we need to save m in n.
R = 1:n;

fprintf('== ACC begins: initial %d clusters, cutoff %.2f\n', nCluster, param.ACC_cutoff);
fprintf('               ');
if bPair, fprintf('iter-image matching ');
else fprintf('intra-image matching '); end
if bPair, fprintf('with intra-cluster 1-to-1');
else fprintf('without any constraints'); end
fprintf('\n');

iterOuter = 0;
while nCluster > 1 
 	iterOuter = iterOuter + 1;
    
    [minValue, k] = min(cur_dist_mat);
    i = floor(m+1/2-sqrt(m^2-m+1/4-2*(k-1)));
    j = k - (i-1)*(m-i/2)+i;
    %new_cluster = { cur_clusters{i}{1:length(cur_clusters{i})} cur_clusters{j}{1:length(cur_clusters{j})} }; % our new cluster
        
    if minValue > param.ACC_cutoff,    break;    end
    
    if param.bOneToOne && cur_ovlp_mat(k)         
        subSubOvlpMatrix = cdata.overlapMatrix(clusterInfo(i).matchIdx, clusterInfo(j).matchIdx);
        overlap1 = any(subSubOvlpMatrix,2);
        overlap2 = any(subSubOvlpMatrix,1);
        n1 = size(subSubOvlpMatrix,1);
        n2 = size(subSubOvlpMatrix,2);
        if n1 > n2
           matchIdxInI = clusterInfo(i).matchIdx;
           matchIdxInJ = clusterInfo(j).matchIdx(find(~overlap2));
           flipOfMatchInI = clusterInfo(i).flipOfMatch;
           flipOfMatchInJ = clusterInfo(j).flipOfMatch(find(~overlap2));
        elseif n1 < n2
           matchIdxInI = clusterInfo(i).matchIdx(find(~overlap1));
           matchIdxInJ = clusterInfo(j).matchIdx;
           flipOfMatchInI = clusterInfo(i).flipOfMatch(find(~overlap1));
           flipOfMatchInJ = clusterInfo(j).flipOfMatch;
        else
            error('error! n1 == n2 selected in confliction');
        end
    else
        matchIdxInI = clusterInfo(i).matchIdx;
        matchIdxInJ = clusterInfo(j).matchIdx;
        flipOfMatchInI = clusterInfo(i).flipOfMatch;
        flipOfMatchInJ = clusterInfo(j).flipOfMatch;
    end
    szClusterI = length(matchIdxInI);    
    szClusterJ = length(matchIdxInJ);
    
    if bPair
        bMergingAnA = 1;
        flipConsistency = 0;
    else
        % check whether merging A-A or A-B 
        subFlip = cdata.flipMatrix(matchIdxInI, matchIdxInJ);
        subFlip(find(flipOfMatchInI),:) = ~subFlip(find(flipOfMatchInI),:);
        subFlip(:,find(flipOfMatchInJ)) = ~subFlip(:,find(flipOfMatchInJ));
        flipConsistency = sum(sum(subFlip))/(szClusterI*szClusterJ);
        bMergingAnA = flipConsistency < 0.5;
    end

    % merge into the larger cluster
    newClusterInfo = clusterInfo(i);
    newClusterInfo.matchIdx = [ matchIdxInI matchIdxInJ ];
    newClusterInfo.size = szClusterI + szClusterJ; 
    if bMergingAnA
        newClusterInfo.flipOfMatch = [ flipOfMatchInI flipOfMatchInJ ];
    else
        % flip matches in matchInfo
        newClusterInfo.flipOfMatch = [ flipOfMatchInI ~flipOfMatchInJ ];
    end
      
    % Update cur_dist_mat. In order to vectorize the computation, we need to compute
    % all the indices corresponding to cluster i and j in cur_dist_mat, denoted by I and J.
    I1 = 1:(i-1); I2 = (i+1):(j-1); I3 = (j+1):m; % these are temp variables
    U = [I1 I2 I3]; 
    % vectorized index of (i,:)
    I = [I1.*(m-(I1+1)/2)-m+i i*(m-(i+1)/2)-m+I2 i*(m-(i+1)/2)-m+I3];
    % vectorized index of (j,:)
    J = [I1.*(m-(I1+1)/2)-m+j I2.*(m-(I2+1)/2)-m+j j*(m-(j+1)/2)-m+I3];
    
    % Update the overlap matrix 
    if param.bOneToOne
        cur_ovlp_mat(I) = cur_ovlp_mat(I) | cur_ovlp_mat(J);
    end
    % Update the dissimilarity matrix
    for iter=1:length(U)
        % Firstly, update min dist matrix
        cur_mindist_mat(I(iter)) = min(cur_mindist_mat(I(iter)), cur_mindist_mat(J(iter)));
        % When min dist between two clusters is greater than the threshold,
        % ignore it!
        if cur_mindist_mat(I(iter)) > param.ACC_cutoff
            cur_dist_mat(I(iter)) = Inf;
            continue;
        end
        
       if param.bOneToOne && cur_ovlp_mat(I(iter)) % depriving-merge
            subSubOvlpMatrix = cdata.overlapMatrix(newClusterInfo.matchIdx, clusterInfo(U(iter)).matchIdx);
            overlapNew = any(subSubOvlpMatrix,2);
            overlapU = any(subSubOvlpMatrix,1);
            nNew = length(newClusterInfo.matchIdx);
            nU = length(clusterInfo(U(iter)).matchIdx);
            if nNew > nU
               elementIdxNew = newClusterInfo.matchIdx;
               elementIdxU = clusterInfo(U(iter)).matchIdx(find(~overlapU));
            elseif nNew < nU
               elementIdxNew = newClusterInfo.matchIdx(find(~overlapNew));
               elementIdxU = clusterInfo(U(iter)).matchIdx;
            else
                elementIdxNew = [];
                elementIdxU = [];
            end
        else
            elementIdxNew = newClusterInfo.matchIdx;
            elementIdxU = clusterInfo(U(iter)).matchIdx;
       end
       szClusterNew = length(elementIdxNew);
       szClusterU = length(elementIdxU); 
       dissim = Inf;

       % compute the inter-cluster dissimilarity
       nCombi = szClusterNew * szClusterU;
       if nCombi > 0
            % compute the adaptive partial linkage
            if nCombi <= param.ACC_kNNLinkage
                kNum = nCombi;
            else
                kNum = max( param.ACC_kNNLinkage, ceil( param.ACC_rNNLinkage * nCombi ) );
            end
            if nCombi <= kNum && ~( param.bOneToOne && ( cur_ovlp_mat(k) || cur_ovlp_mat(I(iter)) ) )
                dissim = ( cur_dist_mat(I(iter))*szClusterI + cur_dist_mat(J(iter))*szClusterJ )/(szClusterI+szClusterJ);
            else
                subSubDistMatrix = cdata.affinityMatrix(elementIdxNew,elementIdxU);
                values = sort( subSubDistMatrix(:), 'ascend' );
                dissim = mean( values(1:kNum) );
            end
            
       end
       cur_dist_mat(I(iter)) = dissim;
   end

   J = [J i*(m-(i+1)/2)-m+j];
   cur_dist_mat(J) = []; % no need for the cluster information about j.
   cur_mindist_mat(J) = []; % no need for the cluster information about j.
   if param.bOneToOne % enduce the matching constraint
       cur_ovlp_mat(J) = [];
   end
    
    % update cur_clusters
    clusterInfo(i) = newClusterInfo;
    clusterInfo(j) = [];
    nCluster = nCluster - 1;
    
   % update m, N, R
   m = m-1;
   N(n+iterOuter) = N(R(i)) + N(R(j));
   R(i) = n+iterOuter;
   R(j:(n-1))=R((j+1):n);

   fprintf('.');   if mod(iterOuter,50) == 0, fprintf('%4d iters\n', iterOuter); end
end
fprintf(':%3d iters \n-> %4d final clusters \n',iterOuter,nCluster);
cdata.nCluster = nCluster;
cdata.clusterInfo = clusterInfo;

