%% MATLAB demo code for Agglomerative Correspondence Clustering V1.2
% This code is based on our ICCV 2009 algorithm and is extended by adopting intra-cluster 1-to-1 contraints
% 
% Please cite our work if you find this code useful in your research. 
% Minsu Cho, Jungmin Lee, and Kyoung Mu Lee, 
% Feature Correspondence and Deformable Object Matching via Agglomerative Correspondence Clustering, 
% Proc. International Conference on Computer Vision (ICCV), 2009.
%
% written by Minsu Cho & Jungmin Lee, July 18, 2011, Seoul National University, Korea
% http://cv.snu.ac.kr/~minsucho/        chominsu@gmail.com
% http://cv.snu.ac.kr/~jungminlee/      jmlee.vision@gmail.com
%
% In order to help the users modify and apply initial feature matching and our algorithm,
% this script contains all the detailed params of each step.
% You can alter feature extraction and initial matching using these params. 
% But, main control params are aparam.ACC_cutoff and aparam.thres_area
% 
% This script saves the results of initial matching and loads it automatically
% in next runs of the same images. If you want to make new initial matching, 
% delete the data file of 'fi_image1+image2.mat' or disable the related lines.
%
% For more detail, refer to README.txt

%% for input and output data
%fileName1 = 'face_1.png';  fileName2 = 'face_2.png';
fileName1 = 'mickey_1.png'; fileName2 = 'mickey_2.png';

iparam.imageFolder = pwd;
iparam.dataSaveFolder = pwd;
iparam.bShow = 1;  % show detected features and initial matches 
                   % when # of features is large, this can take a long time
% set paths
addpath('./commonFunctions/');
addpath('./Algorithms/ACC/');
addpath('./BruteSearch120909');

%% Params for test algorithms

method_to_run = 1;                % 1:ACC (proposed),  0:UNN (Lowe's unambiguous NN matching, a baseline) 
aparam.bOneToOne = 1;             % enforcing the (intra-cluster) one-to-one constraint
% ACC algorithm params
aparam.ACC_cutoff = 30;           % ACC cutoff of dissimilarity 
aparam.ACC_kNNLinkage = 10;       % AP-link kNN value 
aparam.ACC_rNNLinkage = 0.05;     % AP-link ratio (if rNNLinkage == 0 then, single-link)
% UNN algorithm params
aparam.UNN_distRatio = 0.8;       % distance ratio threshold used in UNN matching

% Params for cluster validation
aparam.thres_size = 5;            % cluster reliability threshold (cluster size) 
aparam.thres_area = 0.001;        % cluster reliability threshold (cluster area) = min(areaHull/areaImg in img1,areaHull/areaImg in img2)
aparam.bObjectOverlapCheck = 1;   % eliminate small clusters that overlaps larger clusters (redundant clusters)
                                  % usually, this is not required unless the images have repetitive patterns                                  
aparam.thres_objectOverlap = 0.4; % overlap decision ratio; The clusters overlaps if in both images, overlapArea/largerArea > the ratio



%% Param for feature extraction
fparam.featureScale = 2.0;        % scaling factor for descriptor
fparam.patchSize = 31;            % patch size for descriptor
fparam.validMargin = 10;          % ignore the features in this margin area of image
fparam.bEstimateOrientation = 1;  % estimate the dominant orientation for each feature or not
                                  % when all objects appears in the same direction, set 0
fparam.nMaxOri = 3;               % maximum num of estimated dominant orientations for each feature
fparam.featExt = { 'mser', 'haraff', 'hesaff', 'sift' }; % feature types possible
fparam.bFeatExtUse = [ 1 0 0 0 ]; % feature types used, mark 1s for use

fparam.MSER_Ellipse_Scale = 1.0;                 % MSER 
fparam.MSER_Maximum_Relative_Area = 0.015;        % MSER 
fparam.MSER_Minimum_Size_Of_Output_Region = 25;  % MSER 
fparam.MSER_Minimum_Margin = 5; % or 10          % MSER  
fparam.MSER_Use_Relative_Margins = 0;            % MSER 
fparam.MSER_Vervose_Output = 0;                  % MSER 
fparam.HARAFF_harThres = 5000;                   % Harris-Affine
fparam.HESAFF_hesThres = 500;                    % Hessian-Affine

%% initial matching params for ACC
mparam.nMaxMatch = 1000;                   % max num of initial matches
mparam.bMatchDistribution = 1;             % 1: best of max num 2: equally distrubuted for each feat type
mparam.kNN = 5;                            % max num of NN matches for EACH feature
mparam.distThres = 0.6;                    % threshold value of SIFT distance, when not using, set 0  
mparam.distRatio = 0.0;                    % use Lowe's unambiguous NN; when not using, set 0
mparam.thresholdScaleDiff = 10;            % eliminate matches with large scale diff
mparam.selfmatching_dist_thres = 0.03;     % in the case of initial matching within a sigle image, 
                                           % prevent self-matching by distance ratio w.r.t image size
mparam.bReflective = 0;                    % enable reflective matching
mparam.bFilterMatch = 1;                   % filter out overlapping matches (very similar, redundant matchs)
mparam.redundancy_thres = 3.0;             % criterion to filter out overlapping matches (pixel distance)

%% save file name and paths
iparam.nView = 2;   iparam.bPair = 1;
iparam.view(1).fileName = fileName1(1:end-4);
iparam.view(1).filePathName = [ iparam.imageFolder '/' fileName1];
iparam.view(2).fileName = fileName2(1:end-4);
iparam.view(2).filePathName = [ iparam.imageFolder '/' fileName2];

%% make or load INITIAL matches
if iparam.bPair,    dataTag = [ iparam.view(1).fileName '+' iparam.view(2).fileName ];
else    dataTag = [ iparam.view(1).fileName ];  end        
initPathnFile = [ iparam.dataSaveFolder '/' 'fi_' dataTag '.mat' ];
if exist( initPathnFile ) == 2 % if the file exists
    disp ([ initPathnFile ' file loading.']);
    load(initPathnFile);
else % if the file doesn't exist, perform initial matching routine
    cdata = initialmatch_main( iparam, fparam, mparam );
    fprintf('- computing an affine symmetric transfer errors of %d by %d combination pairs...\n',cdata.nInitialMatches,cdata.nInitialMatches);
    % caculate affinity matrix of initial matches by reprojection error
    tic;
    [ cdata.affinityMatrix cdata.flipMatrix ] = computeAffineTransferDistanceMEX( cdata.view,...
         cell2mat({ cdata.matchInfo(1:cdata.nInitialMatches).match}'), 0 );
    fprintf('>>> affinity matrix calculation done for %d initial matches: %f sec\n', cdata.nInitialMatches,toc);
    cdata.GT = [];
    save (initPathnFile, 'cdata');                                % disable this line if you don't want to save this
    disp([ initPathnFile ' file saved for initial information']); % disable this line if you don't want to save this
end
   
%% Perform MATCHING
if method_to_run == 1      % Agglomerative Correspondence Clustering ICCV2009
    [status, cdata, runningTime ] = wrapper_ACC( aparam, cdata );    algTag = 'ACC';
elseif method_to_run == 0  % Lowe's Unambiguous NN, ICCV1999, for comparison purpose
    % note that this function performs UNN on the features extracted at the previous step,
    % ignoring the initial matches obtained at the previous routine
    [status, cdata, runningTime ] = wrapper_UNN( aparam, cdata );    algTag = 'UNN';
end
runningTime        
%% Validate the clusters (eliminate unreliable clusters)
cdata = validateCluster(cdata, aparam);
nCluster = length(cdata.clusterInfo);
nValidCluster = 0; nValidMatch=0;
for k=1:nCluster
    if cdata.clusterInfo(k).valid, 
        nValidCluster = nValidCluster + 1;
        nValidMatch = nValidMatch + length(cdata.clusterInfo(k).matchIdx);    
    end
end

%% Show the results
hFig0 = figure('Name','input images');
display_inputimages( [], cdata);
scrsz = get(0,'ScreenSize');    set(hFig0, 'Position',[1 scrsz(4)/3 scrsz(3)/1.7 scrsz(4)/1.8]); hold off;
hFig1 = figure('Name','feature matching result'); 
display_result_matching_vis1( [], cdata);
str_out = sprintf('== %s : %d clusters, %d matches', algTag, nValidCluster, nValidMatch);     
title(str_out);
scrsz = get(0,'ScreenSize');    set(hFig1, 'Position',[1 scrsz(4)/3 scrsz(3)/1.7 scrsz(4)/1.8]); hold off;

